#import "DirectoryBrowserCell.h"
#import "../DataSources/DirectoryItem.h"
#import "../Utils/Utils.h"


@implementation DirectoryBrowserCell


@synthesize image		= mImage;
@synthesize numImages	= mNumImages;


+(DirectoryBrowserCell *)cellWithItem:(DirectoryItem *)item
{
	DirectoryBrowserCell * cell = [[[DirectoryBrowserCell alloc] init] autorelease];
	[cell setImage:[[NSWorkspace sharedWorkspace] iconForFile:[item path]]];
	[cell setNumImages:[item numImages]];
	[cell setTitle:[[item path] lastPathComponent]];
	return cell;
}

-(id)copyWithZone:(NSZone *)zone
{
	// NOTE: the "Analyse" option of XCode might report a potential leak at the
	//		 end of the function. This is normal : instead of copying the image
	//		 object, we simply retain it. In the dealloc method, it's released,
	//		 so in practice, it doesn't leak.
	DirectoryBrowserCell * result = (DirectoryBrowserCell *)[super copyWithZone:zone];
	if (result != nil)
	{
		[[result image] retain];
	}
	return result;
}

-(id)init
{
	self = [super init];
	if (self != nil)
	{
		mImage				= nil;
		mNumImages			= 0;
	}
	return self;
}

-(void)dealloc
{
	[mImage release];
	[super dealloc];
}

/**
	This method is called when the view needs to draw an item.
	The drawing is done in 3 parts : first, the icon, second, the number of
	images in the directory (a background and the string) and lastly, the name
	of the directory, which uses the default method.
*/
-(void)drawWithFrame:(NSRect)cellFrame inView:(NSView *)controlView
{
	// draw the image
	if (mImage != nil)
	{
		// create the image frame and update the cellFrame
		NSRect imageFrame;
		NSDivideRect(cellFrame, &imageFrame, &cellFrame,
					 cellFrame.size.height, NSMinXEdge);

		// add a little border to the image frame
		imageFrame.size.height -= 2;
		imageFrame.size.width -= 2;
		imageFrame.origin.y += 1;
		imageFrame.origin.x += 1;

		// offset it to the right, and update the cellFrame width
		imageFrame.origin.x += 3;
		cellFrame.origin.x += 3;
		cellFrame.size.width -= 6;

		// draw the icon
		[mImage drawInRect:imageFrame
				fromRect:NSZeroRect
				operation:NSCompositeSourceOver
				fraction:1.0f
				respectFlipped:YES
				hints:nil];
	}

	// create the string containing the number of images
	NSString * numImages = [NSString stringWithFormat:@"%d", mNumImages];

	// create the attributes for drawing that string (we want it right aligned)
	NSMutableDictionary * attributes = [[NSMutableDictionary alloc] init];
	NSMutableParagraphStyle * paragraphStyle = [Utils defaultParagraphStyle];
	[paragraphStyle setAlignment:NSRightTextAlignment];
	[attributes setObject:paragraphStyle forKey:NSParagraphStyleAttributeName];
	
	// compute the string frame
	NSSize stringSize = [Utils stringSize:numImages withAttribute:attributes];

	// get the frame that we'll use for the background of the number of images
	NSRect frame;
	NSDivideRect(cellFrame, &frame, &cellFrame,
				 stringSize.width + cellFrame.size.height, NSMaxXEdge);

	// shrink the frame a little bit and draw it
	frame.size.height -= 4;
	frame.size.width -= 4;
	frame.origin.y += 2;
	frame.origin.x += 2;
	[[NSColor controlHighlightColor] setStroke];
	[[NSColor controlHighlightColor] setFill];
	NSBezierPath * background = [NSBezierPath bezierPathWithRoundedRect:frame
											  xRadius:frame.size.height / 2
											  yRadius:frame.size.height / 2];
	[background stroke];
	[background fill];

	// offset the frame for the drawing of the number of images
	frame.size.width -= frame.size.height / 2;
	frame.origin.y -= 1;
	frame.origin.x += 1;

	// draw the number of images
	[numImages drawInRect:frame withAttributes:attributes];
	[attributes release];

	// draw the rest
	[super drawWithFrame:cellFrame inView:controlView];
}

@end
