#import "ImageView.h"
#import "../DataSources/ImageDataSource.h"
#import "../DataSources/ImageItem.h"
#import "../Preferences/Preferences.h"
#import "../Utils/FileUtils.h"
#import "../Utils/Utils.h"
#import "../Utils/SlideShow.h"
#import <QuartzCore/QuartzCore.h>


// TODO: implement zooming
#define MIN_ZOOM_FACTOR		0.1f
#define MAX_ZOOM_FACTOR		5.0f


@implementation ImageView


@synthesize dataSource = mDataSource;


-(void)awakeFromNib
{
	DEBUG_LOG(@"ImageView - awakeFromNib");

	mImageURL			= nil;
	mImageIndex			= -1;
	
	// NOTE: When adding this class to your XIB, you can either drop an
	//		 IKImageView, or a simple NSView to your interface, and then change
	//		 the class type to ImageView. The "normal" way of doing this would
	//		 of course be to drop an IKImageView, so that your can configure the
	//		 IKImageView through the interface builder. BUT ! If we do that, you
	//		 will have loooots of bugs, especially with overlay layers for
	//		 animated gifs. So, instead, I just use an NSView. And as a result,
	//		 I have to configure it programmatically here ...
	// NOTE: we need to setImageWithURL once ... It seems this method initialize
	//		 a lot of stuff in the IKImageView, and if we don't do that, the
	//		 following setBackgrondColor, etc. will have not effect until the
	//		 first setImageWithURL call occurs.

	[super setImageWithURL:nil];
	[self setBackgroundColor:[NSColor blackColor]];
	[self setAutoresizes:YES];
}

-(void)dealloc
{
	if (mImageURL != nil)
	{
		[mImageURL release];
		mImageURL = nil;
	}
	[super dealloc];
}

/**
	Simply return the path of the current image.
*/
-(NSString *)currentImage
{
	return [mImageURL path];
}

/**
	setCurrentImage is only called by external classes, each time the currently
	selected image changes.
*/
-(void)setCurrentImage:(NSString *)image
{
	DEBUG_LOG(@"ImageView - setCurrentImage:%@", image);
	
	// we update the view, and cache the image index.
	if (image != nil)
	{
		[self setImageWithURL:[NSURL fileURLWithPath:image] force:NO];
		mImageIndex = [[self dataSource] indexOfImage:image];
	}
	else
	{
		[self setImageWithURL:nil force:NO];
		mImageIndex = -1;
		[self leaveFullscreen];
	}
}

/**
	The method used to update the content of the view. This is where we detect
	animated gifs and load them into an animation layer so that the IKImageView
	can display it correctly.
*/
-(void)setImageWithURL:(NSURL *)url force:(BOOL)force
{
	if (url == mImageURL && force == NO)
		return;

	DEBUG_LOG(@"ImageView - setImageWithURL:%@ force:%@",
			 [url path],
			 force == YES ? @"YES" : @"NO");

	// re-set autoresize to YES (a previous zooming might have set
	// this to NO)
	[self setAutoresizes:YES];

	// checks if a layer is already set and remove all previous animations
	if ([self overlayForType:IKOverlayTypeImage] == nil)
		[self setOverlay:[CALayer layer] forType:IKOverlayTypeImage];
	[[self overlayForType:IKOverlayTypeImage] removeAllAnimations];

	// backup the image url
	if (url != nil)
	{
		if (mImageURL != url)
		{
			[mImageURL release];
			mImageURL = [url copy];
		}

		// set the image of the IKImageView. For a reason I don't know, if we
		// don't do that, the overlay layers will often crash ... Maybe this
		// method initialize some stuff in the IKImageView or whatever ...
		[super setImageWithURL:mImageURL];

		// check if the image a gif
		if ([FileUtils isGIF:[url path]] == YES)
		{
			// create the image
			NSImage * image = [[NSImage alloc] initWithContentsOfFile:[url path]];
			
			// get the image representations
			NSArray * reps = [image representations];
			for (NSImageRep * rep in reps)
			{
				// find the bitmap representation
				if ([rep isKindOfClass:[NSBitmapImageRep class]] == NO)
					continue;

				// get the bitmap representation
				NSBitmapImageRep * bitmapRep = (NSBitmapImageRep *)rep;
				
				// check multiframe gif
				int numFrame = [[bitmapRep valueForProperty:NSImageFrameCount] intValue];
				if (numFrame == 0)
					break;
				
				// create the layer, animation and value array
				CAKeyframeAnimation * animation = [CAKeyframeAnimation animationWithKeyPath:@"contents"];
				NSMutableArray * values = [NSMutableArray array];
				
				// loop through the frames
				float animationDuration = 0.0f;
				for (int i = 0; i < numFrame; ++i)
				{
					// set the current frame
					[bitmapRep setProperty:NSImageCurrentFrame withValue:[NSNumber numberWithInt:i]];
					
					// if the frame duration is 0, replace it by a default of 1/20 second
					if ([[bitmapRep valueForProperty:NSImageCurrentFrameDuration] floatValue] <= 0.000001f)
						[bitmapRep setProperty:NSImageCurrentFrameDuration withValue:[NSNumber numberWithFloat:1.0f / 20.0f]];
					
					// add the CGImageRef to this frame to the value array
					[values addObject:(id)[bitmapRep CGImage]];
					
					// update the duration of the animation
					animationDuration += [[bitmapRep valueForProperty:NSImageCurrentFrameDuration] floatValue];
				}
				
				// setup the animation
				[animation setValues:values];
				[animation setCalculationMode:@"discrete"];
				[animation setDuration:animationDuration];
				[animation setRepeatCount:HUGE_VAL];

				// create the layer and add the animation
				CALayer * layer = [CALayer layer];
				[layer addAnimation:animation forKey:@"contents"];
				[self setOverlay:layer forType:IKOverlayTypeImage];

				// we have our animated gif, no need to continue
				break;
			}

			// release the image
			[image release];
		}
	}
	else if (mImageURL != nil)
	{
		if (mImageURL != nil)
		{
			[mImageURL release];
			mImageURL = nil;
		}
		[super setImageWithURL:nil];
	}
}

-(BOOL)fullscreen
{
	return [self isInFullScreenMode];
}

-(void)setFullscreen:(BOOL)fullscreen
{
	if (fullscreen == YES)
		[self enterFullscreen];
	else
		[self leaveFullscreen];
}

/**
	Enter full screen, and set a few options, such as hiding the cursor, etc.
	It is also important to note that when entering full screen, the image view
	automatically becomes the first responder, and don't let go until it leaves
	fullscreen
*/
-(void)enterFullscreen
{
	if ([self isInFullScreenMode] == YES)
		return;

	DEBUG_LOG(@"ImageView - Enter fullscreen");

	// create a dictionary for the fullscreen options
	// NSFullScreenModeWindowLevel is used to display on top of everything
	// NSApplicationPresentationDefault is used so that we can still use Spaces, swtich process, etc.
	NSMutableDictionary * options = [NSMutableDictionary dictionary];
	[options setValue:[NSNumber numberWithInt:NSMainMenuWindowLevel + 1]
			 forKey:NSFullScreenModeWindowLevel];
	[options setValue:[NSNumber numberWithInt:NSApplicationPresentationDefault]
			 forKey:NSFullScreenModeApplicationPresentationOptions];
	
	// enter fullscreen mode and hide cursor
	[self enterFullScreenMode:[NSScreen mainScreen] withOptions:options];
	[NSCursor hide];
	
	// NOTE: when the image view is not in fullscreen, the image is quite small,
	//		 and when we go to fullscreen, instead of displaying the full
	//		 resolution image, the view only scales the current image, resulting
	//		 in low resolution view (if we change image, the problem doesn't
	//		 happen since the view loads the image at the correct resolution)
	//		 This is why we need to reload the image when we go to fullscreen,
	//		 in order to have a resolution adapted to the fullscreen resolution.
	[self reloadImage];

	// request first responder
//	[[self window] makeFirstResponder:self];
}

/**
	Leave fullscreen. Unhide the cursor, and let go of first responder
*/
-(void)leaveFullscreen
{
	if ([self isInFullScreenMode] == NO)
		return;

	DEBUG_LOG(@"ImageView - Leave fullscreen");

	[self exitFullScreenModeWithOptions:nil];
	[NSCursor unhide];
	[self setAutoresizes:YES];

	// let go of first responder (we can set the nextKeyView to whatever we
	// want in the interface builder. It's currently set to the image browser
	// view)
	[[self window] makeFirstResponder:[self nextKeyView]];
	
	// stop slide show if there's one running
	[SlideShow stopSlideShow];
}

/**
	When not in fullscreen, the IKImageView maintains a small view of the image
	internally. When we go to fullscreen, the resolution of the image doesn't
	change so we have a really poor resolution.
	This method is used to force reloading the image to get a resolution adapted
	to fullscreen.
*/
-(void)reloadImage
{
	DEBUG_LOG(@"ImageView - reload image");

	[self setImageWithURL:mImageURL force:YES];
}


-(BOOL)acceptsFirstResponder
{
	// only accepts being first responder when in fullscreen. When not,
	// the directory and image browsers only should be first responders.
	if ([self isInFullScreenMode] == NO)
	{
		DEBUG_LOG(@"ImageView - becomeFirstResponder - refuses");
		return NO;
	}

	DEBUG_LOG(@"ImageView - becomeFirstResponder - accepts");
	return YES;
}

-(BOOL)resignFirstResponder
{
	// only resign if we are NOT in fullscreen
	if ([self isInFullScreenMode] == YES)
	{
		DEBUG_LOG(@"ImageView - still in fullscreen, don't resign first responder !");
		return NO;
	}

	DEBUG_LOG(@"ImageView - resigning first responder");
	return YES;
}

-(void)selectNextImage:(NSTimer *)timer
{
	[self changeImage:1 loop:[[timer userInfo] boolValue]];
}

/**
	Utility method used to change the current image.
*/
-(void)changeImage:(int)delta loop:(BOOL)loop
{
	int numImages = (int)[[self dataSource] numberOfItemsInImageBrowser:nil];
	int newIndex = mImageIndex + delta;
	if (newIndex < 0)
	{
		if (loop == YES)
		{
			while (newIndex < 0)
				newIndex += numImages;
		}
		else
		{
			newIndex = 0;
		}
	}
	else if (newIndex >= numImages)
	{
		if (loop == YES)
		{
			newIndex = newIndex % numImages;
		}
		else
		{
			newIndex = numImages - 1;
		}
	}

	DEBUG_LOG(@"ImageView - changeImage:[%d -> %d]", mImageIndex, newIndex);

	if (newIndex != mImageIndex)
	{
		mImageIndex = newIndex;
		ImageItem * item = [[self dataSource] imageBrowser:nil
											  itemAtIndex:mImageIndex];
		[self setCurrentImage:[item path]];
	}
}

-(void)keyDown:(NSEvent *)theEvent
{
	// only handle events when in fullscreen
	if ([self isInFullScreenMode] == NO)
		return;

	// get the event and the modifiers
	NSString * characters = [theEvent charactersIgnoringModifiers];
	unichar event = [characters characterAtIndex:0];
	
	DEBUG_LOG(@"ImageView - keyDown - event [%d]", event);
	
	switch (event)
	{
		case ' ':
		case 'p':
			if ([SlideShow isRunning] == YES)
				[SlideShow stopSlideShow];
			else
				[SlideShow startSlideShow:self callback:@"selectNextImage:"];
			break;

		// up & left : advance backward
		case NSUpArrowFunctionKey:
		case NSLeftArrowFunctionKey:
			[self changeImage:-1 loop:NO];
			break;

		// down & right : advance forward
		case NSDownArrowFunctionKey:
		case NSRightArrowFunctionKey:
			[self changeImage:1 loop:NO];
			break;

		// enter & escape : leave fullscreen
		case 13:
		case 27:
			[self setFullscreen:NO];
			break;

		default:
			[super keyDown:theEvent];
	}
}

-(void)mouseDown:(NSEvent *)theEvent
{
	// only handle events when in fullscreen
	if ([self isInFullScreenMode] == NO)
		return;

	DEBUG_LOG(@"ImageView - mouseDown - button [%d] clickCount[%d]",
			  (int)[theEvent buttonNumber],
			  (int)[theEvent clickCount]);

	if ([theEvent clickCount] > 1)
		[self setFullscreen:NO];
}

-(void)otherMouseDown:(NSEvent *)theEvent
{
	// only handle events when in fullscreen
	if ([self isInFullScreenMode] == NO)
		return;

	DEBUG_LOG(@"ImageView - otherMouseDown - button [%d] clickCount[%d]",
			  (int)[theEvent buttonNumber],
			  (int)[theEvent clickCount]);
	
	if ([theEvent clickCount] == 1)
		[self setFullscreen:NO];	
}

-(void)scrollWheel:(NSEvent *)theEvent
{
	// only handle events when in fullscreen
	if ([self isInFullScreenMode] == NO)
		return;

	DEBUG_LOG(@"ImageView - scrollWheel - delta [%.4f] precise [%@]",
			 [theEvent deltaY],
			 [theEvent hasPreciseScrollingDeltas] == YES ? @"YES" : @"NO");

	switch ([theEvent type])
	{
		case NSScrollWheel:
		{
			float delta = [theEvent scrollingDeltaY];
			if (delta > 0.0f)
				[self changeImage:-1 loop:NO];
			else if (delta < 0.0f)
				[self changeImage:1 loop:NO];
			break;
		}
	}
}

@end
